#!/usr/bin/env python
# coding: utf-8

# # 정기후원자의 이탈 예측모형 구현

# 작성일: 2021-08-05

# 작성자: 공학전문대학원 응용공학과 | 김민창 (2021-23150) |  minchang.kim@snu.ac.kr

# ## Install packages

# 기본 통계 패키지를 설치합니다.

# In[1]:


import pandas as pd
import numpy as np
import missingno as msno
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats as stats
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import warnings
warnings.filterwarnings('ignore')


# 기계학습 패키지를 설치합니다.

# In[2]:


from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import AdaBoostClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split
from sklearn.model_selection import GridSearchCV
from sklearn import metrics
from sklearn.metrics import accuracy_score
from sklearn.metrics import roc_curve
from sklearn.metrics import roc_auc_score
from sklearn.metrics import recall_score, confusion_matrix, precision_score, f1_score, accuracy_score, classification_report
from xgboost import XGBClassifier
from lightgbm import LGBMClassifier


# ## Gather Data

# CSV 형식으로 저장된 정기후원자의 데이터를 불러옵니다.

# In[3]:


df = pd.read_csv('./donus_churn_data_v3.csv')


# In[4]:


df.info()


# ## Preprocessing

# Null 값이 존재하는 레코드를 제거합니다.

# In[5]:


df.dropna(subset=['age'], inplace=True)


# 학습에 이용하지 않는 불필요한 Feature를 제거합니다.

# In[6]:


df = df.drop(['member_code'], axis=1)


# In[7]:


df = df.drop(['is_payment_2021-05', 'is_payment_2021-04', 'is_payment_2021-03',
              'is_payment_2021-02', 'is_payment_2021-01', 'is_payment_2020-12', 'is_payment_2020-11',
              'is_payment_2020-10', 'is_payment_2020-09', 'is_payment_2020-08', 'is_payment_2020-07'], axis=1)


# In[8]:


df = df.drop(['general_interactions_2020-07_2021-06'], axis = 1)


# In[9]:


df = df.drop(['is_receipt'], axis=1)


# Data type을 int로 변경합니다.

# In[10]:


df = df.astype({"age": int})


# Preprocessing 결과를 확인합니다.

# In[11]:


df.info()


# In[12]:


df.isnull().sum()


# ## EDA

# Feature Engineering을 진행하기 주요 데이터의 분포와 특징을 파악합니다.

# In[13]:


df['is_payment_2021-06'].value_counts()


# In[14]:


df['payment_method'].value_counts()


# In[15]:


fig = px.histogram(df, x="payment_method", color="is_payment_2021-06", barmode="group", 
                   title="<b>Churn Distribution by Payments<b>")
fig.update_layout(width=700, height=500, bargap=0.1)
fig.show()


# In[16]:


num_cols = ['duration_month_until_2021-05', 'total_payments', 'monthly_amount', 
            'total_amount', 'total_promises', 'total_engagements', 'total_logins']


# In[17]:


def box_plots(feature, df):
    plt.figure(figsize=(7, 2))
    plt.title("Box Plot for {}".format(feature))
    sns.boxplot(df[feature])
    plt.show()

for feat in num_cols:
    box_plots(feat, df)


# In[18]:


plt.figure(figsize=(25, 10))
corr = df.apply(lambda x: pd.factorize(x)[0]).corr()
mask = np.triu(np.ones_like(corr, dtype=bool))
ax = sns.heatmap(corr, mask=mask, xticklabels=corr.columns, yticklabels=corr.columns, 
                 annot=True, linewidths=.2, cmap='coolwarm', vmin=-1, vmax=1)


# ## Feature Engineering

# Churn은 1, Retetion은 0으로 변환합니다.

# In[19]:


def object_to_int(dataframe_series):
    if dataframe_series.dtype == 'object':
        dataframe_series = dataframe_series.eq('no').mul(1)
    return dataframe_series


# In[20]:


df = df.apply(lambda x: object_to_int(x))


# 범주형 데이터는 One Hot Encoding 형식으로 변환합니다.

# In[21]:


df_dummy_payment_method = pd.get_dummies(df['payment_method'])
df = df.join(df_dummy_payment_method.add_prefix('p_method_'))
df = df.drop(['payment_method'], axis = 1)


# In[22]:


df_dummy_sex = pd.get_dummies(df['sex'])
df = df.join(df_dummy_sex.add_prefix('sex_'))
df = df.drop(['sex'], axis = 1)


# In[23]:


df.info()


# 연속형 범주를 가진 데이터의 Outliers를 제거합니다.

# In[24]:


def modify_payments(x):
    if x >= 82: return 82
    else: return x

df['total_payments'] = df['total_payments'].apply(modify_payments)


# In[25]:


def modify_monthly_amount(x):
    if x >= 100000: return 100000
    else: return x

df['monthly_amount'] = df['monthly_amount'].apply(modify_monthly_amount)


# In[26]:


def modify_total_amount(x):
    if x >= 2500000: return 2500000
    else: return x

df['total_amount'] = df['total_amount'].apply(modify_total_amount)


# In[27]:


def modify_promises(x):
    if x >= 6: return 6
    else: return x

df['total_promises'] = df['total_promises'].apply(modify_promises)


# In[28]:


def modify_engagements(x):
    if x >= 3: return 3
    else: return x

df['total_engagements'] = df['total_engagements'].apply(modify_engagements)


# In[29]:


def modify_logins(x):
    if x >= 10: return 10
    else: return x

df['total_logins'] = df['total_logins'].apply(modify_logins)


# In[30]:


def box_plots(feature, df):
    plt.figure(figsize=(7, 2))
    plt.title("Box Plot for {}".format(feature))
    sns.boxplot(df[feature])
    plt.show()

for feat in num_cols:
    box_plots(feat, df)


# In[31]:


def distplot(feature, frame, color='r'):
    plt.figure(figsize=(6,3))
    plt.title("Distribution for {}".format(feature))
    ax = sns.distplot(frame[feature], color=color)


# In[32]:


num_cols = ['duration_month_until_2021-05', 'total_payments', 'monthly_amount', 
            'total_amount', 'total_promises', 'total_engagements', 'total_logins']


# In[33]:


for feat in num_cols: distplot(feat, df)


# ## Split train and test set

# 학습을 진행하기 전 Train set과 Test set을 구분합니다.

# In[34]:


X = df.drop(columns = ['is_payment_2021-06'])
y = df['is_payment_2021-06'].values


# In[35]:


X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.30, random_state=40, stratify=y)


# 연속형 범주를 가진 데이터에 대한 표준화를 수행합니다.

# In[36]:


scaler = StandardScaler()
X_train[num_cols] = scaler.fit_transform(X_train[num_cols])
X_test[num_cols] = scaler.transform(X_test[num_cols])


# In[37]:


df_std = pd.DataFrame(StandardScaler().fit_transform(df[num_cols].astype('float64')), columns=num_cols)
for feat in num_cols: distplot(feat, df_std, color='c')


# ## Prediction

# 기계학습 Classifier로 학습을 진행한 후 Chrun 예측결과를 비교합니다. Chrun을 1, Retention은 0으로 출력됩니다. Churn의 비중이 낮은 데이터셋의 특징을 고려하여 Churn(1)의 f1-score 값을 성능을 평가하기 위한 비교 지표로 설정하였습니다.

# ### KNN

# In[38]:


knn_model = KNeighborsClassifier(n_neighbors = 5) 
knn_model.fit(X_train,y_train)
predicted_y = knn_model.predict(X_test)
accuracy_knn = knn_model.score(X_test,y_test)
print("KNN accuracy:",accuracy_knn)


# In[39]:


print(classification_report(y_test, predicted_y))


# In[40]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, predicted_y),
                annot=True,fmt = "d",linecolor="k",linewidths=3)    
plt.title("KNN CONFUSION MATRIX",fontsize=14)
plt.show()


# ### SVM

# In[41]:


svc_model = SVC(random_state = 1)
svc_model.fit(X_train,y_train)
predict_y = svc_model.predict(X_test)
accuracy_svc = svc_model.score(X_test,y_test)
print("SVM accuracy is :",accuracy_svc)


# In[42]:


print(classification_report(y_test, predict_y))


# In[43]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, predict_y),
                annot=True,fmt = "d",linecolor="k",linewidths=3)    
plt.title("SVM CONFUSION MATRIX",fontsize=14)
plt.show()


# ### Random Forest

# In[44]:


model_rf = RandomForestClassifier(n_estimators = 500 , oob_score = True, n_jobs = -1,
                                  random_state = 50, max_features = 'auto',
                                  class_weight = 'balanced')
model_rf.fit(X_train, y_train)
prediction_test = model_rf.predict(X_test)
print (metrics.accuracy_score(y_test, prediction_test))


# In[45]:


print(classification_report(y_test, prediction_test))


# In[46]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, prediction_test),
                annot=True,fmt = "d",linecolor="k",linewidths=3)    
plt.title("RANDOM FOREST CONFUSION MATRIX",fontsize=14)
plt.show()


# ### Logistic Regression

# In[47]:


lr_model = LogisticRegression()
lr_model.fit(X_train,y_train)
accuracy_lr = lr_model.score(X_test,y_test)
print("Logistic Regression accuracy is :",accuracy_lr)


# In[48]:


lr_pred= lr_model.predict(X_test)
report = classification_report(y_test,lr_pred)
print(report)


# In[49]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, lr_pred),
                annot=True,fmt = "d",linecolor="k",linewidths=3)    
plt.title("LOGISTIC REGRESSION CONFUSION MATRIX",fontsize=14)
plt.show()


# ### Decision Tree Classifier

# In[50]:


dt_model = DecisionTreeClassifier()
dt_model.fit(X_train,y_train)
predictdt_y = dt_model.predict(X_test)
accuracy_dt = dt_model.score(X_test,y_test)
print("Decision Tree accuracy is :",accuracy_dt)


# In[51]:


print(classification_report(y_test, predictdt_y))


# In[52]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, predictdt_y),
                annot=True, fmt = "d",linecolor="k",linewidths=3)    
plt.title("Decesion Tree CONFUSION MATRIX",fontsize=14)
plt.show()


# ### AdaBoost Classifier

# In[53]:


a_model = AdaBoostClassifier()
a_model.fit(X_train,y_train)
a_preds = a_model.predict(X_test)
print("AdaBoost Classifier accuracy")
metrics.accuracy_score(y_test, a_preds)


# In[54]:


print(classification_report(y_test, a_preds))


# In[55]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, a_preds),
                annot=True,fmt = "d",linecolor="k",linewidths=3)
    
plt.title("AdaBoost Classifier Confusion Matrix",fontsize=14)
plt.show()


# ### Gradient Boosting Classifier

# In[56]:


gb = GradientBoostingClassifier()
gb.fit(X_train, y_train)
gb_pred = gb.predict(X_test)
print("Gradient Boosting Classifier", accuracy_score(y_test, gb_pred))


# In[57]:


print(classification_report(y_test, gb_pred))


# In[58]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, gb_pred),
                annot=True,fmt = "d",linecolor="k",linewidths=3)
    
plt.title("Gradient Boosting Classifier Confusion Matrix",fontsize=14)
plt.show()


# ## Prediction (Additional)

# IAB 최종발표 후의 피드백을 반영하여 LightGBM Classifier로 추가 예측을 진행하였습니다. 적용 결과 LightGBM Classifier을 통해 가장 높은 f1-score를 얻을 수 있었습니다.

# ### LightGBM Classifier

# Scikit Learn 라이브러리의 GridSearchCV 함수를 이용하여 AUC를 기준으로 Hyperparameters 최적화를 진행하였습니다. 

# In[ ]:


lgbm = LGBMClassifier(n_estimators=200)
evals = [(X_train, y_train), (X_test, y_test)]
params = {'num_leaves': [100, 200],
          'learning_rate': [0.05, 0.04],
          'subsample': [0.6, 0.8], 
          'max_depth': [100, 200],
          'min_child_samples': [50, 100]}

gridcv = GridSearchCV(lgbm, param_grid=params, cv=3)
gridcv.fit(X_train, y_train, early_stopping_rounds=30, eval_metric="auc", eval_set=evals)
lgbm_roc_score = roc_auc_score(y_test, gridcv.predict_proba(X_test)[:, 1], average='macro')

print('GridSearchCV Best Parameters:', gridcv.best_params_)
print('ROC AUC: {0:.4f}'.format(lgbm_roc_score))


# 최적화를 통해 도출된 Hyperparameters를 적용하여 예측을 수행하였습니다.

# In[59]:


lgbm = LGBMClassifier(n_estimators=1000, learning_rate=0.04, num_leaves=100,
                      max_depth=100, min_child_samples=50, subsample=0.6)
evals = [(X_test, y_test)]
lgbm.fit(X_train, y_train, early_stopping_rounds=100, eval_metric="auc", eval_set=evals, verbose=True)
lgbm_pred = lgbm.predict(X_test)
lgbm_pred_proba = lgbm.predict_proba(X_test)[:, 1]


# In[60]:


print(classification_report(y_test, lgbm_pred))


# In[61]:


plt.figure(figsize=(6,5))
sns.heatmap(confusion_matrix(y_test, lgbm_pred),
                annot=True,fmt = "d",linecolor="k",linewidths=3)    
plt.title("LightGBM Classifier Confusion Matrix",fontsize=14)
plt.show()


# In[62]:


fpr, tpr, thresholds = roc_curve(y_test, lgbm_pred_proba)
plt.plot([0, 1], [0, 1], 'k--' )
plt.plot(fpr, tpr, label='LightBGM Classifier',color = "r")
plt.xlabel('False Positive Rate')
plt.ylabel('True Positive Rate')
plt.title('LightBGM Classifier ROC Curve',fontsize=16)
plt.show();

